use std::{fs, time::Duration};

use anyhow::{bail, Context};
use serde::Deserialize;
use serde_with::{serde_as, DurationSeconds};

use crate::util::validate_freq;

const DEFAULT_FREQUENCY: u32 = 430_500_000;

const fn default_frequency() -> u32 {
    DEFAULT_FREQUENCY
}

const fn default_enable_tx() -> bool {
    true
}

#[derive(Deserialize, Clone)]
pub struct FelinetConfig {
    pub address: String,
}

#[serde_as]
#[derive(Deserialize, Clone)]
pub struct BeaconConfig {
    #[serde_as(as = "DurationSeconds<u64>")]
    pub period_seconds: Duration,
    #[serde(default)]
    pub max_hops: u8,
    pub latitude: Option<f64>,
    pub longitude: Option<f64>,
    pub altitude: Option<f64>,
    pub comment: Option<String>,
    pub antenna_height: Option<u8>,
    pub antenna_gain: Option<f32>,
    pub tx_power: Option<f32>,
}

#[derive(Deserialize, Clone)]
pub struct Config {
    pub callsign: String,
    pub ssid: u8,
    #[serde(default)]
    pub icon: u16,
    #[serde(default = "default_frequency")]
    pub frequency: u32,
    #[serde(default = "default_enable_tx")]
    pub enable_tx: bool,
    pub felinet: Option<FelinetConfig>,
    pub beacon: BeaconConfig,
}

impl Config {
    pub fn load() -> anyhow::Result<Self> {
        let file_contents = read_conf_file()?;
        let config: Self = toml::from_str(&file_contents).context("parsing config file")?;

        if config.callsign == "CHANGEME" {
            bail!("Callsign is not configured in config file");
        }

        if !validate_freq(config.frequency) {
            bail!(
                "Frequency {:.3} MHz out of range (420MHz - 450MHz)",
                config.frequency as f64 / 1_000_000.0
            );
        }

        Ok(config)
    }
}

fn read_conf_file() -> anyhow::Result<String> {
    let filenames = ["config.toml", "/etc/cats_igate/config.toml"];
    for filename in filenames.iter() {
        if let Ok(x) = fs::read_to_string(filename) {
            return Ok(x);
        }
    }

    bail!("Could not open config file. Should either be `config.toml` in the current directory, or in /etc/cats_igate/")
}
